/* 
 * PPRacer 
 * Copyright (C) 2004 Volker Stroebel <volker@planetpenguin.de>
 *
 * Copyright (C) 1999-2001 Jasmin F. Patry
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */

#include "fonts.h"
#include "tcl_util.h"
#include "render_util.h"

#include <string>
#include <list>
#include <map>

//static bool initialized = false;

//static hash_table_t font_table;
static std::map<std::string,font_node_t> fontTable;

//static hash_table_t binding_table;
static std::map<std::string,font_t> bindingTable;


void init_fonts() 
{
/* if (!initialized) {
	//font_table = create_hash_table();
	//binding_table = create_hash_table();
    }
    initialized = true;
*/
} 

static std::map<std::string,font_node_t>::iterator get_font( const char *fontname)
{
    return fontTable.find(fontname);		
}

bool get_font_binding( const char *binding, font_t &font )
{
	std::map<std::string,font_t>::iterator it;
    if((it=bindingTable.find(binding))!=bindingTable.end()){
		font = it->second;
		return true;
	}
	return false;
}

bool load_font( const char *fontname, const char *filename, const char *texname )
{
    font_node_t *fontnode;
    tex_font_metrics_t *tfm;
    texture_node_t *tex;

    print_debug(DEBUG_FONT, "Loading font %s from file: %s", 
		fontname, filename);

	tex = get_texture( texname );
		
    if ( tex == NULL ){
		print_warning( IMPORTANT_WARNING, 
		       "Texture `%s' does not exist", texname );
		return false;
    }

    tfm = load_tex_font_metrics( filename );

    if ( tfm == NULL ) {
    	print_warning( IMPORTANT_WARNING, 
		       "couldn't load font file %s", filename );
    	return false;
    }

	std::map<std::string,font_node_t>::iterator it;
	if ( (it = fontTable.find(fontname))!=fontTable.end() ){
		fontnode = &(it->second);
		print_debug( DEBUG_FONT, "Font %s already exists, deleting...", 
		     fontname );
		delete_tex_font_metrics( fontnode->tfm );
		fontnode->tex.ref_count -= 1;
	}else{
		fontnode = &fontTable[fontname];
		fontnode->ref_count = 0;
    }
	
    fontnode->tfm = tfm;
    fontnode->tex = *tex;
    tex->ref_count++;

    return true;
} 

static bool del_font( char *fontname )
{
    std::map<std::string,font_node_t>::iterator it;
	if ( (it = fontTable.find(fontname))!=fontTable.end() ){
		check_assertion( it->second.ref_count == 0,
			 "Trying to delete font with non-zero reference "
			 "count" );
		delete_tex_font_metrics( it->second.tfm );
		fontTable.erase(it);	
		return true;		
	}
    return false;
}

bool bind_font( const char *binding, const char *fontname, double size, 
		  pp::Color color )
{
    std::map<std::string,font_node_t>::iterator fontit;
	std::map<std::string,font_t>::iterator bindit;
	font_t *font;
	
    print_debug(DEBUG_FONT, "Binding %s to font name: %s", 
		binding, fontname);

    if ((fontit = get_font(fontname))== fontTable.end()) {
		check_assertion(0, "Attempt to bind to unloaded font");
		return false;
    }

	if ((bindit=bindingTable.find(binding))!=bindingTable.end()){
		font = &(bindit->second);
		font->node->ref_count--;
	}else{
		font = &bindingTable[binding];
	}

    font->node = &(fontit->second);
    font->size = size;
    font->color = color;

    font->node->ref_count += 1;

    return true;
}

bool unbind_font( char *binding )
{
	std::map<std::string,font_t>::iterator it;
	
	if((it=bindingTable.find(binding))!=bindingTable.end()){
		it->second.node->ref_count--;
		bindingTable.erase(it);
		return true;		
	}
	return false;
}

bool flush_fonts(void)
{
	std::list<char*> deleteList;
    char *key;
  
	std::map<std::string,font_node_t>::iterator fontit;
	for(fontit=fontTable.begin(); fontit!=fontTable.end(); fontit++){
		if (fontit->second.ref_count == 0) {
	    	deleteList.push_back(key);
		}
	}
	
	std::list<char*>::iterator it;
	
	for(it=deleteList.begin(); it!=deleteList.end(); it++){
		del_font(*it);
	}

    return true;
}

static double get_scale_factor( font_t *font ) 
{
    char *empty_string = "";
    int dummy1, dummy2, max_ascent;

    get_tex_font_string_bbox( font->node->tfm, empty_string, &dummy1, 
			      &max_ascent, &dummy2 );

    return font->size / max_ascent;
}

static void start_font_draw( font_t *font )
{
    double scale_fact = get_scale_factor( font );
    glPushMatrix();
    glScalef( scale_fact,
	      scale_fact,
	      scale_fact );

    glColor4dv( (double*) &font->color );
}

static void end_font_draw( font_t *font )
{
    glPopMatrix();
}

void bind_font_texture( font_t *font )
{
    glBindTexture( GL_TEXTURE_2D, font->node->tex.texture_id );
}

static void font_advance( font_t *font, const char *string )
{
    int w, a, d;
    get_font_metrics( font, string, &w, &a, &d );
    glTranslatef( w, 0, 0 );
}

void draw_character( font_t *font, char c)
{
    char buff[2];

    start_font_draw( font );
    draw_tex_font_char( font->node->tfm, c );
    end_font_draw( font );

    buff[0] = c;
    buff[1] = '\0';

    font_advance( font, buff );
}

void draw_string( font_t *font, const char *string )
{
    start_font_draw( font );
    draw_tex_font_string( font->node->tfm, string );
    end_font_draw( font );

    font_advance( font, string );
}

void get_font_metrics( font_t *font, const char *string,
		       int *width, int *max_ascent, int *max_descent)
{
    double scale_fact = get_scale_factor( font );
    get_tex_font_string_bbox( font->node->tfm, string, width, 
			      max_ascent, max_descent );

    *width = int(*width * scale_fact);
    *max_ascent = int(*max_ascent * scale_fact);
    *max_descent = int(*max_descent * scale_fact);
}


static int load_font_cb (ClientData cd, Tcl_Interp *ip, 
			 int argc, CONST84 char *argv[]) 
{
    bool error = false;

    const char *fontname = NULL;
    const char *filename = NULL;
    const char *texturename = NULL;
    
    if (argc < 2) {
	error = true;
    }

    NEXT_ARG;

    while ( !error && argc > 0 ) {

	if ( strcmp( "-name", *argv ) == 0 ) {
	    NEXT_ARG;
	    if ( argc == 0 ) {
		error = true;
		break;
	    }
	    fontname = *argv;
	} else if ( strcmp( "-file", *argv ) == 0 ) {
	    NEXT_ARG;
	    if ( argc == 0 ) {
		error = true;
		break;
	    }
	    filename = *argv;
	} else if ( strcmp( "-texture", *argv ) == 0 ) {
	    NEXT_ARG;
	    if ( argc == 0 ) {
		error = true;
		break;
	    }
	    texturename = *argv;
	} else {
	    print_warning( TCL_WARNING, "tux_load_font: unrecognized "
			   "parameter `%s'", *argv );
	}

	NEXT_ARG;
    }

    /* Sanity check on values */
    if ( fontname == NULL || filename == NULL || texturename == NULL ) {
	error = true;
    }

    if ( error ) {
	print_warning( TCL_WARNING, "error in call to tux_load_font" );
	Tcl_AppendResult(
	    ip, 
	    "\nUsage: tux_load_font -name <name> -file <file> "
	    "-texture <name> ",
	    (char *) 0 );
	return TCL_ERROR;
    }

    if (!load_font( fontname, filename, texturename )) {
	print_warning( TCL_WARNING, "Could not load font %s", filename );
	return TCL_ERROR;
    }
    
    return TCL_OK;
}


static int bind_font_cb (ClientData cd, Tcl_Interp *ip, 
			 int argc, CONST84 char *argv[]) 
{
    double tmp_dbl;
    bool error = false;

    CONST84 char *binding = NULL;
    CONST84 char *fontname = NULL;
    pp::Color color = pp::Color::white;
    int size = 30;
    
    if (argc < 2) {
	error = true;
    }

    NEXT_ARG;

    while ( !error && argc > 0 ) {

	if ( strcmp( "-binding", *argv ) == 0 ) {
	    NEXT_ARG;
	    if ( argc == 0 ) {
		error = true;
		break;
	    }
	    binding = *argv;
	} else if ( strcmp( "-font", *argv ) == 0 ) {
	    NEXT_ARG;
	    if ( argc == 0 ) {
		error = true;
		break;
	    }
	    fontname = *argv;
	} else if ( strcmp( "-size", *argv ) == 0 ) {
	    NEXT_ARG;
	    if ( argc == 0 ) {
		error = true;
		break;
	    }
	    if ( Tcl_GetDouble ( ip, *argv, &tmp_dbl ) == TCL_ERROR ) {
		error = true;
		break;
	    }
	    size = int(tmp_dbl);
	} else if ( strcmp( "-color", *argv ) == 0 )
	{
	    NEXT_ARG;
	    if ( argc == 0 ) {
		error = true;
		break;
	    }
	    if ( get_tcl_tuple( ip, *argv, (double*)&color, 4 ) == 
		 TCL_ERROR ) 
	    {
		error = true;
		break;
	    }
	} else {
	    print_warning( TCL_WARNING, "tux_bind_font: unrecognized "
			   "parameter `%s'", *argv );
	}

	NEXT_ARG;
    }

    /* Sanity check on values */
    if ( binding == NULL || fontname == NULL || size < 0 ) {
	error = true;
    }

    if ( error ) {
	print_warning( TCL_WARNING, "error in call to tux_bind_font" );
	Tcl_AppendResult(
	    ip, 
	    "\nUsage: tux_bind_font -binding <name> -font <name> "
	    "[-size <height>] ",
	    (char *) 0 );
	return TCL_ERROR;
    }

    if (!bind_font( binding, fontname, size, color )) {
		print_warning( TCL_WARNING, "Could not bind font %s", binding );
		return TCL_ERROR;
    }
    
    return TCL_OK;
}


void register_font_callbacks( Tcl_Interp *ip )
{
    Tcl_CreateCommand (ip, "tux_load_font",   load_font_cb,   0,0);
    Tcl_CreateCommand (ip, "tux_bind_font",   bind_font_cb,   0,0);
}






Font::Font(char* binding, bool centerX, bool centerY)
{
	m_centerX=centerX;
	m_centerY=centerY;
	get_font_binding( binding, m_data );
	bind_font_texture( &m_data );
}

Font::~Font()
{
}

void
Font::draw(const char* string, pp::Vec2d position)
{
	if( m_centerX||m_centerY ){
		int width, asc, desc;
		get_font_metrics(&m_data, string, &width, &asc, &desc);
		
		if(m_centerX){
			position.x -= (width/2.0);
		}
		if(m_centerY){
			position.y -= (asc/2.0);
		}
	}
		
	glPushMatrix();
	{
		glTranslatef( position.x ,position.y,0);
    	draw_string( &m_data, string );
	}
	glPopMatrix();
}
	
void
Font::draw(std::string& string, pp::Vec2d position)
{
	draw(string.c_str(),position);
}

int
Font::getWidth(const char* string)
{
	int width, asc, desc;
	get_font_metrics(&m_data, string, &width, &asc, &desc);
	return width;
}
